#' compute the variogram
#'
#' @param path               `character` where the simulations are to be found
#' @param simulation_id      `character` the id of the simulation to read
#' @param response_var       `character` the variable whose response we are studying
#' @param factor_vars        `character` vector the factor names for which sensitivity will be estimated
#' @param response_reduce_fn `fn` the function to be used to reduce the response variable to one value (default=max)
#' @param factor_reduce_fn   `list(fn)` the functions to be used to reduce the factors variable to one value
#' @param max_to_take        `integer` the maximum number of simulations to read
#' @return a data.table that contains the pairwise variogram statistics (distances, variance, covariance)
#' @export
#'
#' @import data.table
#' @import readcamp
#' @import doParallel
#' @import foreach
#' @importFrom utils combn
#'
#' @examples
#' \dontrun{
#' test <- variogram(path = "/gpfs/scratch/bsc32/bsc032637//out/", simulation_id = "2027582", response_var = "O3", factor_vars = c("TOL", "NO", "NO2", "solar_zenith_angle"), response_reduce_fn = max, factor_reduce_fn = list(data.table::first, data.table::first, data.table::first, mean))
#' }
variogram <- function(
    path,
    simulation_id,
    response_var,
    factor_vars,
    response_reduce_fn = max,
    factor_reduce_fn,
    max_to_take = 100) {
  stopifnot(inherits(path, "character"), length(path) == 1)
  stopifnot(inherits(simulation_id, "character"), length(simulation_id) == 1)
  stopifnot(inherits(response_var, "character"), length(response_var) == 1)
  stopifnot(inherits(factor_vars, "character"))
  stopifnot(inherits(response_reduce_fn, "function"), length(response_reduce_fn) == 1)
  stopifnot(inherits(factor_reduce_fn, "list"), length(factor_reduce_fn) == length(factor_vars))

  for (fn in factor_reduce_fn) {
    stopifnot(inherits(fn, "function"))
  }

  simulation_path <- paste0(path, "/", simulation_id)
  stopifnot(dir.exists(simulation_path))

  gas_names <- readcamp::list_gas_names(simulation_path)

  gas_species_to_read <- c()
  other_vars_to_read <- c()

  if (response_var %in% gas_names) {
    gas_species_to_read <- c(gas_species_to_read, response_var)
  } else {
    other_vars_to_read <- c(other_vars_to_read, response_var)
  }

  for (factor in factor_vars) {
    if (factor %in% gas_names) {
      gas_species_to_read <- c(gas_species_to_read, factor)
    } else {
      other_vars_to_read <- c(other_vars_to_read, factor)
    }
  }

  print("the following vars will be read from the gas_concentration_array")
  print(gas_species_to_read)

  print("the following vars will be read directly from the netcdf")
  print(other_vars_to_read)

  ### READ THE GAS CONCENTRATIONS
  dataset <- readcamp::get_gas_species_concentrations(
    path = simulation_path,
    species = gas_species_to_read,
    max_to_take = max_to_take
  )


  dataset <- data.table::dcast(
    dataset,
    cell + time ~ species,
    value.var = "gas_species_concentrations"
  )
  data.table::setkeyv(dataset, cols = c("cell", "time"))

  ### READ OTHER VARIABLES
  env_vars <- readcamp::get_var_from_all_cells(simulation_path, other_vars_to_read, max_to_take = max_to_take)
  data.table::setkeyv(env_vars, cols = c("cell", "time"))
  ### MERGE AND REDUCE
  dataset <- dataset[env_vars]

  reduced_dataset <- dataset[
    ,
    .(response = response_reduce_fn(get(response_var))),
    by = c("cell")
  ]
  data.table::setnames(reduced_dataset, "response", response_var)

  to_remove <- c()
  for (i in seq_along(factor_vars)) {
    factor <- factor_vars[i]
    reduce_fn <- factor_reduce_fn[[i]]

    temp <- dataset[
      ,
      .(temp_factor = reduce_fn(get(factor))),
      by = c("cell")
    ]

    # scaling to 0-1 interval
    temp[
      ,
      temp_factor := scale(temp_factor, center = min(temp_factor), scale = (max(temp_factor) - min(temp_factor)))
    ]

    if (any(is.nan(temp[["temp_factor"]]))) {
      warning(paste0(" removing ", factor, " because if contains a NaN values after scaling"))
      to_remove <- c(to_remove, i)
      next
    }

    data.table::setnames(temp, "temp_factor", factor)

    reduced_dataset <- reduced_dataset[
      temp,
      on = c("cell")
    ]
  }

  factor_vars <- factor_vars[-to_remove]
  factor_reduce_fn <- factor_vars[-to_remove]

  xselect <- combn(seq_len(nrow(reduced_dataset)), 2)

  mean_response <- mean(reduced_dataset[[response_var]])

  variogram_dataset <- matrix(nrow = ncol(xselect), ncol = 3 + length(factor_vars))
  col_names <- c("pair" = 1, "variance" = 2, "covariance" = 3)
  for (i in seq_along(factor_vars)) {
    col_names[3 + i] <- 3 + i
    names(col_names)[3 + i] <- paste0("h_", factor_vars[i])
  }

  doParallel::registerDoParallel()
  print(paste0("going to use ", foreach::getDoParWorkers(), " workers"))

  variogram_dataset <- foreach::`%dopar%`(foreach::foreach(
    i = seq_len(ncol(xselect)),
    .final = data.table::rbindlist,
    .inorder = FALSE,
    .errorhandling = "remove"
  ), {
    xa <- xselect[1, i]
    xb <- xselect[2, i]
    pair <- paste0("(", xa, ",", xb, ")")
    variance <- 0.5 * (reduced_dataset[xb, ..response_var] - reduced_dataset[xa, ..response_var])^2
    variance <- as.numeric(variance)
    covariance <- (reduced_dataset[xb, ..response_var] - mean_response) * (reduced_dataset[xa, ..response_var] - mean_response)
    covariance <- as.numeric(covariance)

    result <- data.table("pair" = pair, "variance" = variance, "covariance" = covariance)

    for (factor in factor_vars) {
      h <- abs(reduced_dataset[xa, ..factor] - reduced_dataset[xb, ..factor])
      result[[paste0("h_", factor)]] <- as.numeric(h)
    }
    result
  })

  print("computation done")

  return(variogram_dataset)
}
